with
  ada.calendar,
  text_io,
  ada.integer_text_io,
  ada.long_float_text_io,
  ada.numerics.long_elementary_functions,
  terms,
  nlp;

procedure cablenet is

  use
    ada.calendar,
    text_io,
    ada.integer_text_io;

  subtype fmat3x1 is terms.fmat1 (1 .. 3);

  subtype fmat2x3 is terms.fmat2 (1 .. 2, 1 .. 3);

  type rope_rec is
    record
      young   : long_float;
      area    : long_float;
      density : long_float;
    end record;

  type rope_vec is array (integer range <>) of rope_rec;

  type rope_ptr is access rope_vec;

  type cable_rec is
    record
      state   : integer range 0 .. 2 := 0;
      value   : long_float := 0.0;
      rope    : integer;
      glob    : terms.imat1 (1 .. 2);
      length  : long_float;
      u       : fmat3x1;
    end record;

  type cable_vec is array (integer range <>) of cable_rec;

  type cable_ptr is access cable_vec;

  clock1,clock2                        : time;
  halt                                 : exception;
  optimal                              : boolean;
  nn,ne,n,maxls,minor,maxminor,maxlist : integer;
  lstol,mntol,mnref,maxnorm,f          : long_float;
  index                                : terms.imat2_ptr;
  xo,yo,zo,x,g                         : terms.fmat1_ptr;
  known,force                          : terms.fmat2_ptr;
  rope                                 : rope_ptr;
  cable                                : cable_ptr;

  procedure scan (io : in text_io.file_type) is

    ascii : character;

  begin
    loop
      text_io.get(io,ascii);
      text_io.skip_line(io);
      exit when ascii = ':';
    end loop;
  end scan;

  procedure input is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io;

    e,glob,axis : integer;
    nk,nf,nr,ns : integer;
    io          : file_type;

  begin
    open(io, in_file, "input.txt");

    -- The iterations terminate if the infinity norm of the
    -- gradient becomes less than or equal to MAXNORM, which is
    -- calculated as:

    -- MAXNORM = MNTOL * MNREF

    -- If MNTOL is specified as a non-positive value, it is
    -- replaced by 0.001.

    -- If MNREF is specified as a non-positive value, it is
    -- replaced by the infinite norm of the gradient, evaluated at
    -- the starting point.

    scan(io);
    get(io,lstol);
    get(io,maxls);
    get(io,mntol);
    get(io,mnref);
    get(io,maxminor);
    get(io,maxlist);
    skip_line(io);

    if mntol <= 0.0 then
      mntol := 0.001;
    end if;

    -- nodes
    scan(io);
    get(io,nn);
    skip_line(io);

    -- initiate xo, yo, zo
    xo := new terms.fmat1 (1 .. nn);
    yo := new terms.fmat1 (1 .. nn);
    zo := new terms.fmat1 (1 .. nn);

    skip_line(io);
    for k in 1 .. nn loop
      get(io,glob);
      get(io,xo(glob));
      get(io,yo(glob));
      get(io,zo(glob));
      skip_line(io);
    end loop;

    -- ropes
    scan(io);
    get(io,nr);
    skip_line(io);

    -- initiate rope
    rope := new rope_vec (1 .. nr);

    skip_line(io);
    for k in 1 .. nr loop
      get(io,e);
      get(io,rope(e).young);
      get(io,rope(e).area);
      get(io,rope(e).density);
      skip_line(io);
    end loop;

    -- elements
    scan(io);
    get(io,ne);
    skip_line(io);

    -- initiate cable
    cable := new cable_vec (1 .. ne);

    skip_line(io);
    for k in 1 .. ne loop
      get(io,e);
      get(io,cable(e).glob(1));
      get(io,cable(e).glob(2));
      get(io,cable(e).rope);
      skip_line(io);
    end loop;

    -- state
    scan(io);
    get(io,ns);
    skip_line(io);

    if ns > 0 then
      skip_line(io);
      for k in 1 .. ns loop
        get(io,e);
        get(io,cable(e).state);
        get(io,cable(e).value);
        skip_line(io);
      end loop;
    end if;

    -- initiate known, index
    known := new terms.fmat2 (1 .. nn, 1 .. 3);
    index := new terms.imat2 (1 .. nn, 1 .. 3);

    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        known(glob,axis) := 0.0;
        index(glob,axis) := 0;
      end loop;
    end loop;

    -- known displacements
    scan(io);
    get(io,nk);
    skip_line(io);

    if nk > 0 then
      skip_line(io);
      for k in 1 .. nk loop
        get(io,glob);
        get(io,axis);
        get(io,known(glob,axis));
        skip_line(io);
        index(glob,axis) := 1;
      end loop;
    end if;

    -- initiate force
    force := new terms.fmat2 (1 .. nn, 1 .. 3);

    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        force(glob,axis) := 0.0;
      end loop;
    end loop;

    -- nodal forces
    scan(io);
    get(io,nf);
    skip_line(io);

    if nf > 0 then
      skip_line(io);
      for k in 1 .. nf loop
        get(io,glob);
        get(io,axis);
        get(io,force(glob,axis));
        skip_line(io);
      end loop;
    end if;

    close(io);
  end input;

  procedure cableobj (cable                   : in cable_rec;
                      epsilon,strain,axial,fe : out long_float;
                      x                       : in terms.fmat1_ptr;
                      ge                      : out fmat2x3) is

    use
      ada.numerics.long_elementary_functions;

    glob,i        : integer;
    c,alpha,normw : long_float;
    z,w           : terms.fmat1 (1 .. 3);
    xe            : terms.fmat2 (1 .. 2, 1 .. 3);

  begin
    -- global displacements
    for loc in 1 .. 2 loop
      glob := cable.glob(loc);
      for axis in 1 .. 3 loop
        i := index(glob,axis);
        if i > n then
          xe(loc,axis) := known(glob,axis);
        else
          xe(loc,axis) := x(i);
        end if;
      end loop;
    end loop;

    -- calculate z
    for axis in 1 .. 3 loop
      z(axis) := (xe(2,axis) - xe(1,axis)) / cable.length;
    end loop;

    -- calculate w
    normw := 0.0;
    for axis in 1 .. 3 loop
      w(axis) := cable.u(axis) + z(axis);
      normw := normw + w(axis) ** 2;
    end loop;
    normw := sqrt(normw);

    alpha := 0.0;
    for axis in 1 .. 3 loop
      alpha := alpha + z(axis) * (cable.u(axis) + w(axis));
    end loop;

    epsilon := alpha / (normw + 1.0);

    case cable.state is
      -- normal
      when 0 =>
        strain := epsilon;

        axial := rope(cable.rope).young * strain * rope(cable.rope).area;

        -- energy
        fe := 0.5 * axial * strain * cable.length;

      -- constant cut
      when 1 =>
        c := cable.value / cable.length;

        strain := (epsilon + c) / (1.0 - c);

        axial := rope(cable.rope).young * strain * rope(cable.rope).area;

        -- energy
        fe := 0.5 * axial * strain * (cable.length - cable.value);

      -- constant tension
      when 2 =>
        strain := cable.value / rope(cable.rope).young;

        axial := cable.value * rope(cable.rope).area;

        -- energy
        fe := cable.length * axial * normw;
    end case;

    -- residue
    c := axial / normw;
    for axis in 1 .. 3 loop
      alpha := c * w(axis);
      ge(1,axis) := -alpha;
      ge(2,axis) :=  alpha;
    end loop;
  end cableobj;

  procedure objective (n : in integer;
                       f : out long_float;
                       x : in terms.fmat1_ptr;
                       g : in terms.fmat1_ptr) is

    glob,i                  : integer;
    epsilon,strain,axial,fe : long_float;
    ge                      : fmat2x3;

  begin
    -- initiate objective, gradient
    f := 0.0;
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        i := index(glob,axis);
        if i <= n then
          f := f - force(glob,axis) * x(i);
          g(i) := -force(glob,axis);
        end if;
      end loop;
    end loop;

    -- elements
    for e in 1 .. ne loop
      cableobj(cable(e),epsilon,strain,axial,fe,x,ge);

      -- update objective
      f := f + fe;

      -- update gradient
      for loc in 1 .. 2 loop
        glob := cable(e).glob(loc);
        for axis in 1 .. 3 loop
          i := index(glob,axis);
          if i <= n then
            g(i) := g(i) + ge(loc,axis);
          end if;
        end loop;
      end loop;
    end loop;
  end objective;

  package qn is new nlp.qnewton(objective);

  procedure start is

    use
      ada.numerics.long_elementary_functions;

    glob,glob1,glob2,s1,s2 : integer;
    norm,aux               : long_float;

  begin
    -- number of unknowns
    n := 3 * nn;
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        if index(glob,axis) = 1 then
          n := n - 1;
        end if;
      end loop;
    end loop;

    -- new sequence
    s1 := 0;
    s2 := n;
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        if index(glob,axis) = 0 then
          s1 := s1 + 1;
          index(glob,axis) := s1;
        else
          s2 := s2 + 1;
          index(glob,axis) := s2;
        end if;
      end loop;
    end loop;

    -- elements
    for e in 1 .. ne loop
      glob1 := cable(e).glob(1);
      glob2 := cable(e).glob(2);

      -- u
      cable(e).u(1) := xo(glob2) - xo(glob1);
      cable(e).u(2) := yo(glob2) - yo(glob1);
      cable(e).u(3) := zo(glob2) - zo(glob1);

      norm := 0.0;
      for axis in 1 .. 3 loop
        norm := norm + cable(e).u(axis) ** 2;
      end loop;
      norm := sqrt(norm);

      for axis in 1 .. 3 loop
        cable(e).u(axis) := cable(e).u(axis) / norm;
      end loop;

      cable(e).length := norm;
    end loop;

    -- add self weight for elements
    for e in 1 .. ne loop
      aux := cable(e).length * rope(cable(e).rope).density / 2.0;
      for loc in 1 .. 2 loop
        glob := cable(e).glob(loc);
        force(glob,3) := force(glob,3) - aux;
      end loop;
    end loop;

    -- initiate x
    x := new terms.fmat1 (1 .. n);

    for k in 1 .. n loop
      x(k) := 0.0;
    end loop;

    -- initiate g
    g := new terms.fmat1 (1 .. n);

    -- maxnorm
    if mnref <= 0.0 then
      objective(n,f,x,g);
      maxnorm := mntol * nlp.norm(n,g);
    else
      maxnorm := mntol * mnref;
    end if;
  end start;

  procedure vertex (glob : in integer;
                    vx   : out fmat3x1) is

    i : integer;

  begin
    vx(1) := xo(glob);
    vx(2) := yo(glob);
    vx(3) := zo(glob);

    for axis in 1 .. 3 loop
      i := index(glob,axis);
      if i <= n then
        vx(axis) := vx(axis) + x(i);
      else
        vx(axis) := vx(axis) + known(glob,axis);
      end if;
    end loop;
  end vertex;

  procedure autocad is

    use
      text_io,
      ada.long_float_text_io;

    glob : integer;
    vx   : fmat3x1;
    io   : file_type;

  begin
    create(io, out_file, "autocad.scr");

    -- set running object snap mode to NONE
    put(io,"osmode 0");
    new_line(io);

    -- undeformed, state 0
    put(io,"layer make State_0 ");
    new_line(io);

    for e in 1 .. ne loop
      if cable(e).state = 0 then
        put(io,"3dpoly ");
        for loc in 1 .. 2 loop
          glob := cable(e).glob(loc);
          put(io,xo(glob),1,8,3);
          put(io,",");
          put(io,yo(glob),1,8,3);
          put(io,",");
          put(io,zo(glob),1,8,3);
          put(io," ");
        end loop;
        new_line(io);
      end if;
    end loop;

    -- undeformed, state 1
    put(io,"layer make State_1 ");
    new_line(io);

    for e in 1 .. ne loop
      if cable(e).state = 1 then
        put(io,"3dpoly ");
        for loc in 1 .. 2 loop
          glob := cable(e).glob(loc);
          put(io,xo(glob),1,8,3);
          put(io,",");
          put(io,yo(glob),1,8,3);
          put(io,",");
          put(io,zo(glob),1,8,3);
          put(io," ");
        end loop;
        new_line(io);
      end if;
    end loop;

    -- undeformed, state 2
    put(io,"layer make State_2 ");
    new_line(io);

    for e in 1 .. ne loop
      if cable(e).state = 2 then
        put(io,"3dpoly ");
        for loc in 1 .. 2 loop
          glob := cable(e).glob(loc);
          put(io,xo(glob),1,8,3);
          put(io,",");
          put(io,yo(glob),1,8,3);
          put(io,",");
          put(io,zo(glob),1,8,3);
          put(io," ");
        end loop;
        new_line(io);
      end if;
    end loop;

    -- deformed
    put(io,"layer make Deformed ");
    new_line(io);

    for e in 1 .. ne loop
      put(io,"3dpoly ");
      for loc in 1 .. 2 loop
        glob := cable(e).glob(loc);
        vertex(glob,vx);
        put(io,vx(1),1,8,3);
        put(io,",");
        put(io,vx(2),1,8,3);
        put(io,",");
        put(io,vx(3),1,8,3);
        put(io," ");
      end loop;
      new_line(io);
    end loop;

    put(io,"layer off State_0 ");
    new_line(io);

    put(io,"layer off State_1 ");
    new_line(io);

    put(io,"layer off State_2 ");
    new_line(io);

    put(io,"layer set 0 ");
    new_line(io);

    put(io,"zoom ext ext");
    new_line(io);

    close(io);
  end autocad;

  procedure putchar (io : in file_type;
                     c  : in character;
                     n  : in integer) is

    use
      text_io;

  begin
    for k in 1 .. n loop
      put(io,c);
    end loop;
  end putchar;

  procedure output is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io;

    glob,i,last             : integer;
    epsilon,strain,axial,fe : long_float;
    cut,tension             : long_float;
    vx                      : fmat3x1;
    ge                      : fmat2x3;
    io                      : file_type;

  begin
    create(io, out_file, "output.txt");

    putchar(io,'-',65);
    new_line(io);

    put(io,"| Minor Iterations                            | ");
    put(io,minor,15);
    put(io," |");
    new_line(io);

    put(io,"| Total Potential Energy                      | ");
    put(io,f,2,8,3);
    put(io," |");
    new_line(io);

    put(io,"| Allowable Maximum Residue                   | ");
    put(io,maxnorm,2,8,3);
    put(io," |");
    new_line(io);

    putchar(io,'-',65);
    new_line(io);

    -- displacements
    new_line(io);
    new_line(io);
    putchar(io,'-',65);
    new_line(io);
    put(io," Node             Displ X             Displ Y             Displ Z");
    new_line(io);
    putchar(io,'-',65);
    new_line(io);

    new_line(io);

    for glob in 1 .. nn loop
      put(io,glob,5);
      for axis in 1 .. 3 loop
        putchar(io,' ',5);
        i := index(glob,axis);
        if i > n then
          put(io,known(glob,axis),2,8,3);
        else
          put(io,x(i),2,8,3);
        end if;
      end loop;
      new_line(io);
    end loop;

    -- vertex location
    new_line(io);
    new_line(io);
    putchar(io,'-',65);
    new_line(io);
    put(io," Node             Coord X             Coord Y             Coord Z");
    new_line(io);
    putchar(io,'-',65);
    new_line(io);

    new_line(io);

    for glob in 1 .. nn loop
      vertex(glob,vx);
      put(io,glob,5);
      for axis in 1 .. 3 loop
        putchar(io,' ',5);
        put(io,vx(axis),2,8,3);
      end loop;
      new_line(io);
    end loop;

    -- initiate support reactions
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        force(glob,axis) := -force(glob,axis);
      end loop;
    end loop;

    -- axial load
    new_line(io);
    new_line(io);
    putchar(io,'-',27);
    new_line(io);
    put(io,"Element          Axial Load");
    new_line(io);
    putchar(io,'-',27);
    new_line(io);

    new_line(io);

    for e in 1 .. ne loop
      cableobj(cable(e),epsilon,strain,axial,fe,x,ge);

      put(io,e,7);
      putchar(io,' ',5);
      put(io,axial,2,8,3);
      new_line(io);

      -- update support reactions
      for loc in 1 .. 2 loop
        glob := cable(e).glob(loc);
        for axis in 1 .. 3 loop
          force(glob,axis) := force(glob,axis) + ge(loc,axis);
        end loop;
      end loop;
    end loop;

    -- equivalence between constant cut and constant tension
    new_line(io);
    new_line(io);
    putchar(io,'-',57);
    new_line(io);
    put(io,"Element     State                 Cut             Tension");
    new_line(io);
    putchar(io,'-',57);
    new_line(io);

    new_line(io);

    for e in 1 .. ne loop
      case cable(e).state is
        -- normal
        when 0 =>
          null;

        -- constant cut
        when 1 =>
          cableobj(cable(e),epsilon,strain,axial,fe,x,ge);
          tension := rope(cable(e).rope).young * strain;

          put(io,e,7);
          putchar(io,' ',5);
          put(io,cable(e).state,5);
          putchar(io,' ',5);
          put(io,cable(e).value,2,8,3);
          putchar(io,' ',5);
          put(io,tension,2,8,3);
          new_line(io);

        -- constant tension
        when 2 =>
          cableobj(cable(e),epsilon,strain,axial,fe,x,ge);
          cut := cable(e).length / (1.0 + strain) * (strain - epsilon);

          put(io,e,7);
          putchar(io,' ',5);
          put(io,cable(e).state,5);
          putchar(io,' ',5);
          put(io,cut,2,8,3);
          putchar(io,' ',5);
          put(io,cable(e).value,2,8,3);
          new_line(io);
      end case;
    end loop;

    -- support reactions
    new_line(io);
    new_line(io);
    putchar(io,'-',65);
    new_line(io);
    put(io," Node             Force X             Force Y             Force Z");
    new_line(io);
    putchar(io,'-',65);
    new_line(io);

    new_line(io);

    for glob in 1 .. nn loop
      last := 0;
      for axis in 1 .. 3 loop
        if index(glob,axis) > n and abs force(glob,axis) > maxnorm then
          last := axis;
        end if;
      end loop;

      if last /= 0 then
        put(io,glob,5);
        for axis in 1 .. last loop
          if index(glob,axis) > n and abs force(glob,axis) > maxnorm then
            putchar(io,' ',5);
            put(io,force(glob,axis),2,8,3);
          else
            putchar(io,' ',20);
          end if;
        end loop;
        new_line(io);
      end if;
    end loop;

    close(io);
  end output;

begin
  input;

  start;

  clock1 := clock;

  qn.lbfgs(maxlist,n,maxminor,maxls,
           minor,
           maxnorm,lstol,
           f,
           x,g,
           optimal);

  if optimal then
    put("Optimal");
  else
    put("NOT Optimal");
  end if;
  new_line;

  put("Minor = ");
  put(minor,5);
  new_line;

  clock2 := clock;

  new_line;
  put("Elapsed time = ");
  put(integer(clock2 - clock1));
  put(" seconds");
  new_line;

  autocad;

  output;
end cablenet;
